const txt = new TextEncoder;
const toUTF8 = txt.encode.bind(txt);
const ALGORITHM = 'AWS4-HMAC-SHA256';

// https://github.com/aws/aws-sdk-js/blob/cc29728c1c4178969ebabe3bbe6b6f3159436394/lib/signers/v4.js#L190-L198
const NOSIGNING = ['authorization', 'content-type', 'content-length', 'user-agent', 'presigned-expires', 'expect', 'x-amzn-trace-id'];

async function hmac(bits, value) {
	let key = await crypto.subtle.importKey('raw', bits, { name: 'HMAC', hash: 'SHA-256' }, false, ['sign']);
	return crypto.subtle.sign('HMAC', key, toUTF8(value));
}

function hash(msg) {
	return crypto.subtle.digest('SHA-256', toUTF8(msg)).then(toHEX);
}

function toHEX(bits) {
	let i=0, hex='', arr = new Uint8Array(bits);
	for (; i < arr.length; i++) hex += arr[i].toString(16).padStart(2, '0');
	return hex;
}

async function signkey(secret, values=[]) {
	let val, key = toUTF8(`AWS4${secret}`);
	let arr = values.concat('aws4_request');
	for (val of arr) key = await hmac(key, val);
	return [key, arr.join('/')];
}

export default function (clientKey, secretKey, config={}) {
	return async function (req, opts={}) {
		let { method, pathname, query, body, headers={} } = req;

		let service = opts.service || config.service;
		let region = opts.region || config.region;

		let date = headers['X-Amz-Date'] = headers['Date'] = (
			opts.date || (new Date).toUTCString()
		);

		let hkey, hval, headers_signed='', headers_string='';
		for ([hkey, hval] of new Headers(headers)) {
			if (!NOSIGNING.includes(hkey)) {
				headers_string += `${hkey}:${hval}\n`;
				if (headers_signed) headers_signed += ';';
				headers_signed += hkey;
			}
		}

		let reqinfo = await hash(
			[
				method || (body ? 'POST' : 'GET'),
				pathname || '/', query || '',
				headers_string, headers_signed,
				await hash(body || '')
			].join('\n')
		);

		let [signing_key, credential_scope] = await signkey(secretKey, [date, region, service]);
		let signature = await hmac(signing_key, [ALGORITHM, date, credential_scope, reqinfo].join('\n')).then(toHEX);

		return {
			Algorithm: ALGORITHM,
			Credential: `${clientKey}/${credential_scope}`,
			SignedHeaders: headers_signed,
			Signature: signature,
		};
	};
}
